@@ -79,6 +79,27 @@ class ModOntology(Ontology):
     def is_ontology_available(self):
         # TODO will be developed properly in the ontology_helper
         return self._is_ontology_available
+    
+    def is_optional(self, uri):
+        self.lock.acquire()
+        res = super().is_optional(uri)
+        self.lock.release()
+        return res
+    
+    def parse_extra_triples(self, extra_triples, format='n3', parse_oda_annotations = True):
+        self.lock.acquire()
+        super().parse_extra_triples(
+            extra_triples,
+            format=format,
+            parse_oda_annotations=parse_oda_annotations)
+        self.lock.release()
+
+    def get_parameter_hierarchy(self, param_uri):
+        self.lock.acquire()
+        res = super().get_parameter_hierarchy(param_uri)
+        self.lock.release()
+        return res
+    
 
 ontology = ModOntology(oda_ontology_path)
 oda_prefix = str([x[1] for x in ontology.g.namespaces() if x[0] == 'oda'][0])
@@ -390,7 +411,7 @@ class NotebookAdapter:
         for i, x in enumerate(comment_tokens):
             if x.start[0]==l:
                 res = comment_tokens.pop(i)
-                return res.string[1:]
+                return res.string.strip(' #')
         return ''
     
     def parse_source_multiline(self, source: str) -> dict[str, list[dict]]:
@@ -448,7 +469,7 @@ class NotebookAdapter:
             
         # now parse full-line comments
         for comment in comments:
-            cstring = comment.string[1:]
+            cstring = comment.string.strip('# ')
             result['standalone'].append(cstring)        
         
         return result
@@ -820,14 +841,22 @@ class NotebookAdapter:
         adapted_parameters = copy.deepcopy(parameters)
         exceptions = []
         posix_path_with_annotations_pattern = re.compile(rf"^{re.escape(oda_prefix)}.*_POSIXPath_")
+        file_url_with_annotations_pattern = re.compile(rf"^{re.escape(oda_prefix)}.*_FileURL_")
+        file_reference_with_annotations_pattern = re.compile(rf"^{re.escape(oda_prefix)}.*_FileReference_")
         for input_par_name, input_par_obj in self.input_parameters.items():
             if ontology.is_ontology_available:
                 parameter_hierarchy = ontology.get_parameter_hierarchy(input_par_obj['owl_type'])
                 is_posix_path = f"{oda_prefix}POSIXPath" in parameter_hierarchy
+                is_file_url = f"{oda_prefix}FileURL" in parameter_hierarchy
+                is_file_reference = f"{oda_prefix}FileReference" in parameter_hierarchy
             else:
                 is_posix_path = f"{oda_prefix}POSIXPath" == input_par_obj['owl_type'] or \
                                 posix_path_with_annotations_pattern.match(input_par_obj['owl_type']) is not None
-            if is_posix_path:
+                is_file_url = f"{oda_prefix}FileURL" == input_par_obj['owl_type'] or \
+                                file_url_with_annotations_pattern.match(input_par_obj['owl_type']) is not None
+                is_file_reference = f"{oda_prefix}FileReference" == input_par_obj['owl_type'] or \
+                                    file_reference_with_annotations_pattern.match(input_par_obj['owl_type']) is not None
+            if is_posix_path or is_file_url or is_file_reference:
                 arg_par_value = parameters.get(input_par_name, None)
                 if arg_par_value is None:
                     arg_par_value = input_par_obj['default_value']
@@ -838,21 +867,26 @@ class NotebookAdapter:
                         token = context.get('token', None)
                         if token is not None:
                             logger.debug(f'adding token to the url: {arg_par_value}')
-                            url_parts = urlparse(adapted_parameters[input_par_name])
+                            url_to_adapt = adapted_parameters.get(input_par_name, arg_par_value)
+                            url_parts = urlparse(url_to_adapt)
                             url_args = parse_qs(url_parts.query)
-                            url_args['token'] = [token] # the values in the dictionary need to be lists
-                            new_url_parts = url_parts._replace(query=urlencode(url_args, doseq=True))
-                            adapted_parameters[input_par_name] = urlunparse(new_url_parts)
-                            logger.debug(f"updated url: {adapted_parameters[input_par_name]}")
-                            arg_par_value = adapted_parameters[input_par_name]
+                            if token not in url_args:
+                                url_args['token'] = [token] # the values in the dictionary need to be lists
+                                new_url_parts = url_parts._replace(query=urlencode(url_args, doseq=True))
+                                adapted_parameters[input_par_name] = urlunparse(new_url_parts)
+                                logger.debug(f"updated url: {adapted_parameters[input_par_name]}")
+                                arg_par_value = adapted_parameters[input_par_name]
 
                     logger.debug(f'download {arg_par_value}')
                     try:
-                        file_name = self.download_file(arg_par_value, tmpdir)
-                        adapted_parameters[input_par_name] = file_name
+                        if is_posix_path:
+                            file_name = self.download_file(arg_par_value, tmpdir)
+                            adapted_parameters[input_par_name] = file_name
                     except Exception as e:
                         exceptions.append(e)
-
+                elif is_file_url:
+                    exceptions.append(ValueError(
+                        f'Parameter {input_par_name} value "{arg_par_value}" can not be interpreted as FileURL.'))
         return dict(
             adapted_parameters=adapted_parameters,
             exceptions=exceptions
@@ -1093,16 +1127,19 @@ def validate_oda_dispatcher(nba: NotebookAdapter, optional=True, machine_readabl
             logger.error("dispatcher validation is not optional!")
             raise
     else:
-        nbpq = NB2WProductQuery('testname', 
-                        'testproduct', 
-                        nba.extract_parameters(),
-                        nba.extract_output_declarations())
+        nbpq = NB2WProductQuery('testname',
+                                'testproduct',
+                                json.loads(json.dumps(nba.extract_parameters(), cls=CustomJSONEncoder)),
+                                json.loads(json.dumps(nba.extract_output_declarations(), cls=CustomJSONEncoder)),
+                                oda_ontology_path)
 
         output = nba.extract_output()
 
         logger.debug(json.dumps(output, indent=4))
 
         class MockRes:
+            headers={'content-type': 'application/json'}
+
             @staticmethod
             def json():
                 return {
@@ -1111,17 +1148,19 @@ def validate_oda_dispatcher(nba: NotebookAdapter, optional=True, machine_readabl
                     }
                 }
 
-        logger.debug("parameters as interpreted by dispatcher: %s", json.dumps(json.loads(nbpq.get_parameters_list_as_json()), indent=4))
+        logger.debug("parameters as interpreted by dispatcher: %s", json.dumps(nbpq.get_parameters_list_jsonifiable(), indent=4))
 
-        dispatcher_parameters = json.loads(nbpq.get_parameters_list_as_json())
+        dispatcher_parameters = nbpq.get_parameters_list_jsonifiable()
 
         for parameter in dispatcher_parameters:
-            logger.info("\033[32mODA dispatcher parameter \033[0m: %s", parameter)
+            if 'query_name' not in parameter.keys() and 'product_name' not in parameter.keys():
+                logger.info("\033[32mODA dispatcher parameter \033[0m: %s", parameter)
 
         prod_list = nbpq.build_product_list(instrument=None, res=MockRes, out_dir=None)
 
         for prod in prod_list:
-            logger.info("\033[33mworkflow the output produces ODA product \033[0m: \033[31m%s\033[0m (%s) %s", prod.name, prod.type_key, prod)
+            logger.info("\033[33mworkflow the output produces ODA product \033[0m: \033[31m%s\033[0m (%s) %s", 
+                        prod.name, prod.type_key, prod)
 
         if machine_readable:
             print("WORKFLOW-DISPATCHER-SIGNATURE:", json.dumps([
