@@ -9,6 +9,7 @@ try:
 except ImportError:
     from werkzeug.routing import MethodNotAllowed, NotFound
 
+
 import queue
 from nb2workflow import ontology, publish, schedule
 from nb2workflow.nbadapter import NotebookAdapter, find_notebooks, PapermillWorkflowIncomplete
@@ -18,7 +19,6 @@ import json
 import glob
 import time
 import logging
-import inspect
 import requests
 import base64
 import hashlib
@@ -26,17 +26,18 @@ import datetime
 import tempfile
 import nbformat
 import yaml
+import traceback
 
 from io import BytesIO
+from bs4 import BeautifulSoup
 
 
 from flask import Flask, make_response, jsonify, request, url_for, send_file, Response
 from flask_caching import Cache
-from flask_cors import CORS
 
 from flasgger import LazyString, Swagger, swag_from
 
-from nb2workflow.workflows import serialize_workflow_exception
+from nb2workflow.helpers import serialize_workflow_exception
 from nb2workflow.json import CustomJSONEncoder
 
 import threading
@@ -74,9 +75,12 @@ cache = Cache(config={'CACHE_TYPE': 'simple'})
 def create_app():
     app = Flask(__name__)
 
+    app.config['SWAGGER'] = {
+        'uiversion': 3,
+        "openapi": "3.0.3",
+    }
     template = {
         "swaggerUiPrefix": LazyString(lambda: request.environ.get('HTTP_X_FORWARDED_PREFIX', '')),
-        "swagger": "2.0",
         "info": {
             "title": "ODAHub API",
             "description": "",
@@ -93,6 +97,7 @@ def create_app():
     swagger = Swagger(app, template=template)
     app.wsgi_app = ReverseProxied(app.wsgi_app)
     app.json_encoder = CustomJSONEncoder
+    app.config["JSON_SORT_KEYS"] = False
     cache.init_app(app, config={'CACHE_TYPE': 'SimpleCache'})
 
 
@@ -132,19 +137,24 @@ class AsyncWorker(threading.Thread):
         async_workflow.run()
 
 class AsyncWorkflow:
-    def __init__(self, key, target, params, callback=None):
+    def __init__(self, key, target, params, context={}):
         self.key = key
         self.target = target
         self.params = params
-        self.callback = callback
+        self.context = context
+
+        logger.info("%s initializing callback %s", self, self.callback)
+
+    @property
+    def callback(self):
+        return self.context.get('callback', None)
 
-        logger.info("%s initializing callback %s", self, callback)
 
     def run(self):
         try:
             self._run()
         except Exception as e:
-            logger.error("run failed unexplicably: %s", repr(e))
+            logger.error("run failed inexplicably: %s\n%s", repr(e), traceback.format_exc())
             app.async_workflows[self.key] = dict(
                 output={}, 
                 exceptions=[serialize_workflow_exception(e)]
@@ -170,13 +180,19 @@ class AsyncWorkflow:
             return
        
         template_nba = app.notebook_adapters.get(self.target)
-        nba = NotebookAdapter(template_nba.notebook_fn, tempdir_cache=app.async_workflow_jobdirs)
+        nba = NotebookAdapter(template_nba.notebook_fn, tempdir_cache=app.async_workflow_jobdirs,
+                              n_download_max_tries=template_nba.n_download_max_tries,
+                              download_retry_sleep_s=template_nba.download_retry_sleep_s,
+                              max_download_size=template_nba.max_download_size)
         
         app.async_workflows[self.key] = 'started'
         self.perform_callback(action='progress')
 
         try:
-            exceptions = nba.execute(self.params['request_parameters'], callback_url=self.callback, tmpdir_key=self.key)
+            thread_id = threading.get_ident()
+            process_id = os.getpid()
+            logger.info(f'nba.execute thread id: {thread_id} ; process id: {process_id}')
+            exceptions = nba.execute(self.params['request_parameters'], context=self.context, tmpdir_key=self.key)
         except PapermillWorkflowIncomplete as e:
             logger.info("found incomplete workflow: %s, rescheduling", repr(e))
 
@@ -255,6 +271,8 @@ def workflow(target, background=False, async_request=False):
 
     async_request = request.args.get('_async_request', async_request)
     async_request_callback = request.args.get('_async_request_callback', None)
+    token = request.args.get("_token", None)
+    context = dict(callback=async_request_callback, token=token)
 
     logger.debug("target %s", target)
 
@@ -262,7 +280,10 @@ def workflow(target, background=False, async_request=False):
         logger.debug("raw parameters %s", request.args)
 
     template_nba = app.notebook_adapters.get(target)
-    nba = NotebookAdapter(template_nba.notebook_fn)
+    nba = NotebookAdapter(template_nba.notebook_fn,
+                          n_download_max_tries=template_nba.n_download_max_tries,
+                          download_retry_sleep_s=template_nba.download_retry_sleep_s,
+                          max_download_size=template_nba.max_download_size)
 
     if nba is None:
         interpreted_parameters = None
@@ -279,6 +300,13 @@ def workflow(target, background=False, async_request=False):
 
     # async
     if async_request:
+        if len(issues) > 0:
+            exceptions = [serialize_workflow_exception(ValueError(v)) for v in issues]
+            return make_response(jsonify(workflow_status="done",
+                                        data=dict(output="incomplete", 
+                                                    exceptions=exceptions),
+                                        comment=""), 200)
+
         key = hashlib.sha224(json.dumps(
             dict(target=target, params=interpreted_parameters)).encode('utf-8')).hexdigest()
 
@@ -290,7 +318,8 @@ def workflow(target, background=False, async_request=False):
             async_task = AsyncWorkflow(key=key,
                                        target=target,
                                        params=interpreted_parameters,
-                                       callback=async_request_callback)
+                                       context=context
+                                       )
 
             app.async_workflows[key] = 'submitted'
             async_queue.put(async_task)
@@ -319,23 +348,26 @@ def workflow(target, background=False, async_request=False):
     if len(issues) > 0:
         return make_response(jsonify(issues=issues), 400)
     else:
-        exceptions = nba.execute(interpreted_parameters['request_parameters'])
+        exceptions = nba.execute(interpreted_parameters['request_parameters'], context=context)
 
-        nretry = 10
-        while nretry > 0:
-            try:
-                output = nba.extract_output()
-                if len(output) == 0:
+        if not os.path.exists(nba.output_notebook_fn):
+            output = {}
+        else:
+            nretry = 10
+            while nretry > 0:
+                try:
+                    output = nba.extract_output()
+                    if len(output) == 0:
+                        logger.debug(
+                            "output from notebook is empty, something failed, attempts left: %s", nretry)
+                    else:
+                        break
+                except nbformat.reader.NotJSONError as e:
                     logger.debug(
-                        "output from notebook is empty, something failed, attempts left: %s", nretry)
-                else:
-                    break
-            except nbformat.reader.NotJSONError as e:
-                logger.debug(
-                    "output notebook incomplte %s attempts left: %s", e, nretry)
+                        "output notebook incomplte %s attempts left: %s", e, nretry)
 
-            nretry -= 1
-            time.sleep(1)
+                nretry -= 1
+                time.sleep(1)
 
         logger.debug("output: %s", output)
         logger.debug("exceptions: %s", exceptions)
@@ -410,7 +442,12 @@ def setup_routes(app):
                 {
                     "name": p_name,
                     "in": "query",
-                    "type": to_oapi_type(p_data['python_type']),
+                    "schema": {
+                        "type": to_oapi_type(p_data['python_type']),
+                        "nullable": p_data.get('is_optional', False)
+                        # strictly speaking, there is no way to set query parameter to null
+                        # we define a convention in which '%00' string represents null
+                        },
                     "required": False,
                     "default": p_data['default_value'],
                     "description": p_data['comment']+" "+p_data['owl_type'],
@@ -727,6 +764,10 @@ def main():
         for nba_name, nba in app.notebook_adapters.items():
             publish.publish(args.publish, nba_name, publish_host, publish_port)
 
+    thread_id = threading.get_ident()
+    process_id = os.getpid()
+    logger.info(f'service main thread id: {thread_id} ; process id: {process_id}')
+
   #  for rule in app.url_map.iter_rules():
  #       logger.debug("==>> %s %s %s %s",rule,rule.endpoint,rule.__class__,rule.__dict__)
 
@@ -830,6 +871,8 @@ def trace_get_func(job, func):
     if func == "custom.css":
         return ""
 
+    include_glued_output = request.args.get('include_glued_output', True) == 'True'
+
     from nbconvert.exporters import HTMLExporter
     exporter = HTMLExporter()
 
@@ -837,6 +880,15 @@ def trace_get_func(job, func):
 
     output, resources = exporter.from_filename(fn)
 
+    if not include_glued_output:
+        logger.info("include_glued_output arg passed")
+        soup = BeautifulSoup(output, 'html.parser')
+        div_glued_output = soup.find('div', {'class': 'celltag_injected-gather-outputs'})
+        if div_glued_output is not None:
+            div_glued_output.decompose()
+        output = str(soup)
+        logger.info("div element removed form html")
+
     return output
 
 
