@@ -7,7 +7,8 @@ import time
 
 logger = logging.getLogger(__name__)
 
-from urllib.parse import urlencode
+from urllib.parse import urlencode, urlparse, parse_qs
+
 
 @pytest.fixture
 def app():
@@ -39,6 +40,131 @@ def test_posix_download_file_extra_annotations(client):
     r = client.get('/api/v1.0/get/testposixpath_extra_annotated', query_string={'fits_file_path': 'https://fits.gsfc.nasa.gov/samples/testkeys.fits'})
     assert r.json['output']['output_file_download'] == 'file downloaded successfully'
 
+@pytest.mark.parametrize("query_string_fits_file_path", ["generic_url", "mmoda_url", None])
+def test_file_reference(client, query_string_fits_file_path):
+    status_callback_file = "status.json"
+    callback_url = 'file://' + status_callback_file
+    token = 'abc123'
+    query_string = dict(
+        _async_request='no',
+        _async_request_callback=callback_url,
+        _token=token)
+
+    fits_file_path_value = None
+    if query_string_fits_file_path == "generic_url":
+        fits_file_path_value = "https://fits.gsfc.nasa.gov/samples/testkeys.fits"
+    elif query_string_fits_file_path == "mmoda_url":
+        fits_file_path_value = "https://www.astro.unige.ch/mmoda/dispatch-data/test.fits"
+
+    query_string['fits_file_path'] = fits_file_path_value
+
+    r = client.get(f'/api/v1.0/get/testfilereference_extra_annotated', query_string=query_string)
+    assert r.status_code == 201
+
+    from nb2workflow.service import AsyncWorker
+
+    def test_worker_run():
+        AsyncWorker('test-worker').run_one()
+
+    test_worker_thread = threading.Thread(target=test_worker_run)
+    test_worker_thread.start()
+
+    while True:
+        options = client.get('/api/v1.0/options')
+        assert options.status_code == 200
+
+        r = client.get(f'/api/v1.0/get/testfilereference_extra_annotated',
+                       query_string=query_string)
+
+        logger.info('service returns %s %s', r, r.json)
+
+        if r.json['workflow_status'] == 'done':
+            logger.info('workflow done!')
+            break
+
+        time.sleep(0.1)
+
+    test_worker_thread.join()
+    assert 'data' in r.json
+    assert 'output' in r.json['data']
+    assert 'fits_file_path_modified' in r.json['data']['output']
+
+    if query_string_fits_file_path is None:
+        assert r.json['data']['output']['fits_file_path_modified'] == "/home/local/test.fits"
+    elif query_string_fits_file_path == "mmoda_url":
+        url_parts = urlparse(r.json['data']['output']['fits_file_path_modified'])
+        url_args = parse_qs(url_parts.query)
+        assert 'token' in url_args
+    elif query_string_fits_file_path == "generic_url":
+        url_parts = urlparse(r.json['data']['output']['fits_file_path_modified'])
+        url_args = parse_qs(url_parts.query)
+        assert 'token' not in url_args
+
+@pytest.mark.parametrize("query_string_fits_file_url", ["generic_url", "file_path", "numeric", None])
+def test_mmoda_file_url(client, query_string_fits_file_url):
+    status_callback_file = "status.json"
+    callback_url = 'file://' + status_callback_file
+    token = 'abc123'
+    query_string = dict(
+        _async_request='no',
+        _async_request_callback=callback_url,
+        _token=token)
+
+    fits_file_url_value = None
+    if query_string_fits_file_url == "generic_url":
+        fits_file_url_value = "https://fits.gsfc.nasa.gov/samples/testkeys.fits"
+    elif query_string_fits_file_url == "file_path":
+        fits_file_url_value = "/home/local/test.fits"
+    elif query_string_fits_file_url == "numeric":
+        fits_file_url_value = 123456
+
+    query_string['fits_file_url'] = fits_file_url_value
+
+    r = client.get(f'/api/v1.0/get/testfileurl_extra_annotated', query_string=query_string)
+
+    assert r.status_code == 201
+
+    from nb2workflow.service import AsyncWorker
+
+    def test_worker_run():
+        AsyncWorker('test-worker').run_one()
+
+    test_worker_thread = threading.Thread(target=test_worker_run)
+    test_worker_thread.start()
+
+    while True:
+        options = client.get('/api/v1.0/options')
+        assert options.status_code == 200
+
+        r = client.get(f'/api/v1.0/get/testfileurl_extra_annotated',
+                       query_string=query_string)
+
+        logger.info('service returns %s %s', r, r.json)
+
+        if r.json['workflow_status'] == 'done':
+            logger.info('workflow done!')
+            break
+
+        time.sleep(0.1)
+
+    test_worker_thread.join()
+    assert 'data' in r.json
+    assert 'output' in r.json['data']
+    if query_string_fits_file_url != "file_path" and query_string_fits_file_url != "numeric":
+        assert 'mmoda_url_modified' in r.json['data']['output']
+        url_parts = urlparse(r.json['data']['output']['mmoda_url_modified'])
+        url_args = parse_qs(url_parts.query)
+        assert 'token' in url_args
+
+        assert 'fits_file_url_modified' in r.json['data']['output']
+        url_parts = urlparse(r.json['data']['output']['fits_file_url_modified'])
+        url_args = parse_qs(url_parts.query)
+        assert 'token' not in url_args
+
+    else:
+        assert r.json['data']['output'] == 'incomplete'
+        assert r.json['data']['exceptions'][0]['edump'] == f"ValueError('Parameter fits_file_url value \"{fits_file_url_value}\" can not be interpreted as FileURL.')"
+
 def test_posix_download_file_with_arg_low_download_limit(client, app_low_download_limit):
     r = client.get('/api/v1.0/get/testposixpath', query_string={'fits_file_path': 'https://fits.gsfc.nasa.gov/samples/testkeys.fits'})
     assert r.json['output'] == {}
