@@ -1,4 +1,5 @@
 import json
+import os
 from oda_api.data_products import ODAAstropyTable
 from nb2workflow.json import CustomJSONEncoder
 import numpy as np
@@ -9,8 +10,9 @@ import nb2workflow
 
 @pytest.fixture
 def app():
+    testfiles_path = os.path.join(os.path.dirname(__file__), 'testfiles')
     app = nb2workflow.service.app
-    app.notebook_adapters = nb2workflow.nbadapter.find_notebooks('tests/testfiles')
+    app.notebook_adapters = nb2workflow.nbadapter.find_notebooks(testfiles_path)
     nb2workflow.service.setup_routes(app)
     print("creating app")
     return app
